/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*

 */
#ifndef GPSR_H
#define GPSR_H

#include "gpsr-ptable.h"
#include "ns3/node.h"
#include "gpsr-packet.h"
#include "ns3/ipv4-routing-protocol.h"
#include "ns3/ipv4-interface.h"
#include "ns3/ipv4-l3-protocol.h"
#include "ns3/ip-l4-protocol.h"

#include "ns3/mobility-model.h"
#include "gpsr-rqueue.h"
#include "ns3/ipv4-header.h"
#include "ns3/ipv4-address.h"
#include "ns3/ipv4-route.h"
#include "ns3/location-service.h"
#include "ns3/god.h"
#include "src/wifi/model/snr-tag.h"

#include "ns3/random-variable-stream.h"

#include <map>
#include <complex>
#include <utility>

namespace ns3
{
  namespace gpsr
  {
    /**
     * \ingroup gpsr
     *
     * \brief GPSR routing protocol
     */
    class RoutingProtocol : public Ipv4RoutingProtocol
    {
    public:
      static TypeId GetTypeId (void);
      static const uint32_t GPSR_PORT;

      /// c-tor
      RoutingProtocol ();
      virtual ~RoutingProtocol ();
      virtual void DoDispose ();

      ///\name From Ipv4RoutingProtocol
      //
      //
      Ptr<Ipv4Route> RouteOutput (Ptr<Packet> p, const Ipv4Header &header, Ptr<NetDevice> oif, Socket::SocketErrno &sockerr);
      bool RouteInput (Ptr<const Packet> p, const Ipv4Header &header, Ptr<const NetDevice> idev,
		       UnicastForwardCallback ucb, MulticastForwardCallback mcb,
		       LocalDeliverCallback lcb, ErrorCallback ecb);
      virtual void NotifyInterfaceUp (uint32_t interface);
      int GetProtocolNumber (void) const;
      virtual void AddHeaders (Ptr<Packet> p, Ipv4Address source, Ipv4Address destination, uint8_t protocol, Ptr<Ipv4Route> route);
      virtual void NotifyInterfaceDown (uint32_t interface);
      virtual void NotifyAddAddress (uint32_t interface, Ipv4InterfaceAddress address);
      virtual void NotifyRemoveAddress (uint32_t interface, Ipv4InterfaceAddress address);
      virtual void SetIpv4 (Ptr<Ipv4> ipv4);
      virtual void RecvGPSR (Ptr<Socket> socket);
      virtual void UpdateRouteToNeighbor (Ipv4Address sender, Ipv4Address receiver, Vector Pos, Vector Vel, double snr);
      virtual void SendHello ();
      virtual bool IsMyOwnAddress (Ipv4Address src);

      virtual void SendSreq (); // police car only, send surveillance request

      Ptr<Ipv4> m_ipv4;
      /// Raw socket per each IP interface, map socket -> iface address (IP + mask)
      std::map< Ptr<Socket>, Ipv4InterfaceAddress > m_socketAddresses;
      /// Loopback device used to defer RREQ until packet will be fully formed
      Ptr<NetDevice> m_lo;

      Ptr<LocationService> GetLS ();
      void SetLS (Ptr<LocationService> locationService);

      /// Broadcast ID
      uint32_t m_requestId;
      /// Request sequence number
      uint32_t m_seqNo;

      /// Number of RREQs used for RREQ rate control
      uint16_t m_rreqCount;
      Time HelloInterval;

      void SetDownTarget (IpL4Protocol::DownTargetCallback callback);
      IpL4Protocol::DownTargetCallback GetDownTarget (void) const;

      virtual void PrintRoutingTable (ns3::Ptr<ns3::OutputStreamWrapper>) const;
      //  std::string PrintPositionTable ();

      // have we received SREQ? is the zone valid?
      bool ShouldBeStreaming ();
      std::vector<uint32_t> GetDestNodes () { return m_dests; }

      // set the node to be a police node, zone and radius
      void SetZone (uint64_t x, uint64_t y, uint64_t r);
      uint64_t GetRadius ();
      void SetRadius (uint64_t r);

      // static calculation functions, assign after call
      static uint64_t ScaleZoneUp (uint64_t radius,
                                   std::vector<std::pair<uint64_t, uint64_t> > members);
      static uint64_t ScaleZoneDown (uint64_t radius,
                                     std::vector<std::pair<uint64_t, uint64_t> > members);

      void SetPoliceNode ();

      // check if in cluster zone
      bool InCluster ();
      bool InHelloPlusZone ();

      uint8_t GetBufferRemaining ();

      static uint64_t DistanceBetween(uint64_t x1, uint64_t y1, uint64_t x2, uint64_t y2);

      // fetch coordinates of node
      uint32_t GetX ()
      {
        Ptr<MobilityModel> MM = m_ipv4->GetObject<MobilityModel> ();
        return (uint32_t) MM->GetPosition ().x;
      }
      uint32_t GetY ()
      {
        Ptr<MobilityModel> MM = m_ipv4->GetObject<MobilityModel> ();
        return (uint32_t) MM->GetPosition ().y;
      }

    private:
      /// Start protocol operation
      void Start ();
      /// Queue packet and send route request
      void DeferredRouteOutput (Ptr<const Packet> p, const Ipv4Header & header, UnicastForwardCallback ucb, ErrorCallback ecb);
      /// If route exists and valid, forward packet.
      void HelloTimerExpire ();

      /// Queue packet and send route request
      Ptr<Ipv4Route> LoopbackRoute (const Ipv4Header & header, Ptr<NetDevice> oif);

      /// If route exists and valid, forward packet.
      bool Forwarding (Ptr<const Packet> p, const Ipv4Header & header, UnicastForwardCallback ucb, ErrorCallback ecb);

      /// Find socket with local interface address iface
      Ptr<Socket> FindSocketWithInterfaceAddress (Ipv4InterfaceAddress iface) const;

      //Check packet from deffered route output queue and send if position is already available
      //returns true if the IP should be erased from the list (was sent/droped)
      bool SendPacketFromQueue (Ipv4Address dst);

      //Calls SendPacketFromQueue and re-schedules
      void CheckQueue ();

      void RecoveryMode(Ipv4Address dst, Ptr<Packet> p, UnicastForwardCallback ucb, Ipv4Header header);

      uint32_t MaxQueueLen;                  ///< The maximum number of packets that we allow a routing protocol to buffer.
      Time MaxQueueTime;                     ///< The maximum period of time that a routing protocol is allowed to buffer a packet for.
      RequestQueue m_queue;

      Timer HelloIntervalTimer;
      Timer CheckQueueTimer;
      uint8_t LocationServiceName;
      PositionTable m_neighbors;
      bool PerimeterMode;
      std::list<Ipv4Address> m_queuedAddresses;
      Ptr<LocationService> m_locationService;

      IpL4Protocol::DownTargetCallback m_downTarget;

      // new random number generators
      Ptr<UniformRandomVariable> m_randomFirstJitter;
      Ptr<UniformRandomVariable> m_randomJitter;

      // zone data
      int64_t m_zoneCentrex;
      int64_t m_zoneCentrey;
      int64_t m_zoneRadius;
      int64_t m_zoneExpiry;
      uint32_t m_requester;
      std::vector<uint32_t> m_dests;
      std::map<uint32_t, MemberData> m_clusterMembers;

      bool m_policeNode = false; // flag for if the node is a police car

    };
  }
}
#endif /* GPSRROUTINGPROTOCOL_H */
