/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include "gpsr-packet.h"
#include "ns3/address-utils.h"
#include "ns3/packet.h"
#include "ns3/log.h"

NS_LOG_COMPONENT_DEFINE ("GpsrPacket");

namespace ns3
{
  namespace gpsr
  {

    NS_OBJECT_ENSURE_REGISTERED (TypeHeader);

    TypeHeader::TypeHeader(MessageType t = GPSRTYPE_HELLO) :
      m_type(t), m_valid(true)
    {
    }

    TypeId TypeHeader::GetTypeId()
    {
      static TypeId tid =
        TypeId("ns3::gpsr::TypeHeader").SetParent<Header>().AddConstructor<
          TypeHeader>();
      return tid;
    }

    TypeId TypeHeader::GetInstanceTypeId() const
    {
      return GetTypeId();
    }

    uint32_t TypeHeader::GetSerializedSize() const
    {
      return 1;
    }

    void TypeHeader::Serialize(Buffer::Iterator i) const
    {
      i.WriteU8((uint8_t) m_type);
    }

    // TODO
    uint32_t TypeHeader::Deserialize(Buffer::Iterator start)
    {
      Buffer::Iterator i = start;
      uint8_t type = i.ReadU8();
      m_valid = true;
      switch (type)
        {
        case GPSRTYPE_HELLO:
        case GPSRTYPE_POS:
          {
            m_type = (MessageType) type;
            break;
          }
        default:
          m_valid = false;
        }
      uint32_t dist = i.GetDistanceFrom(start);
      NS_ASSERT(dist == GetSerializedSize());
      return dist;
    }

    void TypeHeader::Print(std::ostream &os) const
    {
      switch (m_type)
        {
        case GPSRTYPE_HELLO:
          {
            os << "HELLO";
            break;
          }
        case GPSRTYPE_POS:
          {
            os << "POSITION";
            break;
          }
        default:
          os << "UNKNOWN_TYPE";
        }
    }

    bool TypeHeader::operator==(TypeHeader const & o) const
    {
      return (m_type == o.m_type && m_valid == o.m_valid);
    }

    std::ostream &
    operator<<(std::ostream & os, TypeHeader const & h)
    {
      h.Print(os);
      return os;
    }

    //-----------------------------------------------------------------------------
    // HELLO
    //-----------------------------------------------------------------------------
    HelloHeader::HelloHeader(uint64_t originPosx, uint64_t originPosy,
                             int64_t originVelx, int64_t originVely) :
      m_originPosx(originPosx), m_originPosy(originPosy),
      m_originVelx(originVelx), m_originVely(originVely)

    {
    }

    NS_OBJECT_ENSURE_REGISTERED (HelloHeader);

    TypeId HelloHeader::GetTypeId()
    {
      static TypeId tid =
        TypeId("ns3::gpsr::HelloHeader").SetParent<Header>().AddConstructor<
          HelloHeader>();
      return tid;
    }

    TypeId HelloHeader::GetInstanceTypeId() const
    {
      return GetTypeId();
    }

    uint32_t HelloHeader::GetSerializedSize() const
    {
      return 72 + GetClusterSize () * 29 + m_dests.size () * 4;
    }

    void HelloHeader::Serialize(Buffer::Iterator i) const
    {
      NS_LOG_DEBUG("Serialize X " << m_originPosx << " Y " << m_originPosy);
      NS_LOG_DEBUG("Serialize V.x " << m_originVelx << " V.y " << m_originVely);

      i.WriteHtonU64(m_originPosx);
      i.WriteHtonU64(m_originPosy);
      i.WriteHtonU64(m_originVelx);
      i.WriteHtonU64(m_originVely);

      // serialise the new fields
      i.WriteHtonU64(m_zoneCentrex);
      i.WriteHtonU64(m_zoneCentrey);
      i.WriteHtonU64(m_zoneRadius);
      i.WriteHtonU64(m_zoneExpiry);

      // add dests
      i.WriteU32 (m_dests.size ());
      std::vector<uint32_t>::const_iterator d;
      for(d = m_dests.begin () ; d < m_dests.end (); d++)
        {
          i.WriteHtonU32 (*d);
        }

      i.WriteU32 (GetClusterSize ()); // note that we are including the size
      // std::cout << "serialising "<< GetClusterSize () << " members " << std::endl;
      std::map<uint32_t, MemberData>::const_iterator m;
      for (m = m_clusterMembers.begin (); m != m_clusterMembers.end (); ++m)
        {
          //std::cout << "writing member" << std::endl;
          // ip
          i.WriteHtonU32 ((*m).first);
          // md
          i.WriteHtonU64 ((*m).second.x);
          i.WriteHtonU64 ((*m).second.y);
          i.WriteHtonU64 ((*m).second.lastSeen);
          i.WriteU8 ((*m).second.remainingBufferPercentage);
        }
    }

    uint32_t HelloHeader::Deserialize(Buffer::Iterator start)
    {
      //printf("hello deserialise\n");

      Buffer::Iterator i = start;

      m_originPosx = i.ReadNtohU64();
      m_originPosy = i.ReadNtohU64();
      m_originVelx = i.ReadNtohU64();
      m_originVely = i.ReadNtohU64();

      // deserialise added fields
      m_zoneCentrex = i.ReadNtohU64();
      m_zoneCentrey = i.ReadNtohU64();
      m_zoneRadius = i.ReadNtohU64();
      m_zoneExpiry = i.ReadNtohU64();

      // fetch the desinations
      uint32_t dests = i.ReadU32();
      //m_dests.clear();
      for (uint32_t d = 0; d < dests; d++)
        {
          uint32_t address = i.ReadU32();
          m_dests.push_back (address);
        }


      // get the nodes in the cluster
      uint32_t members = i.ReadU32();//ReadNtohU32();
      //m_clusterMembers.clear();

      for (uint32_t m = 0; m < members; m++)
        {
          uint32_t address = i.ReadU32 ();
          MemberData md;
          md.x = i.ReadU64 ();
          md.y = i.ReadU64 ();
          md.lastSeen = i.ReadU64 ();
          md.remainingBufferPercentage = i.ReadU8 ();
          m_clusterMembers.insert (std::make_pair (address, md));
        }

      // resume standard hello deserialisation
      uint32_t dist = i.GetDistanceFrom(start);
      // printf("hello deserialise\n");

      //std::cout << "Size " << dist;
      //std::cout << " vs  " << GetSerializedSize() << "\n";

      NS_ASSERT(dist == GetSerializedSize());
      return dist;
    }

    void HelloHeader::Print(std::ostream &os) const
    {
      os << " PositionX: " << m_originPosx << " PositionY: " << m_originPosy
         << " VelocityX: " << m_originVelx << " VelocityY: " << m_originVely;
    }

    std::ostream &
    operator<<(std::ostream & os, HelloHeader const & h)
    {
      h.Print(os);
      return os;
    }

    bool
    HelloHeader::operator== (HelloHeader const & o) const
    {
      return (m_originPosx == o.m_originPosx && m_originPosy == o.m_originPosy
              && m_originVelx == o.m_originVelx && m_originVely == o.m_originVely);
    }

    void
    HelloHeader::AddClusterMember (uint32_t ip, MemberData md)
    {
      m_clusterMembers[ip] = md;
      /*
      // doesn't exist, add it
      if (m_clusterMembers.find (ip) == m_clusterMembers.end ())
        {
          m_clusterMembers[ip] = md;
          return;
        }

      // exists, verify ip has been seen more recently by neighbour
      if (md.lastSeen > m_clusterMembers[ip].lastSeen)
        {
          m_clusterMembers[ip] = md;
          }*/
    }

    //-----------------------------------------------------------------------------
    // Position
    //-----------------------------------------------------------------------------
    PositionHeader::PositionHeader(uint64_t dstPosx, uint64_t dstPosy,
                                   uint32_t updated, uint64_t recPosx, uint64_t recPosy, uint8_t inRec,
                                   uint64_t lastPosx, uint64_t lastPosy) :
      m_dstPosx(dstPosx), m_dstPosy(dstPosy), m_updated(updated), m_recPosx(
                                                                            recPosx), m_recPosy(recPosy), m_inRec(inRec), m_lastPosx(
                                                                                                                                     lastPosx), m_lastPosy(lastPosy)
    {
    }

    NS_OBJECT_ENSURE_REGISTERED (PositionHeader);

    TypeId PositionHeader::GetTypeId()
    {
      static TypeId tid =
        TypeId("ns3::gpsr::PositionHeader").SetParent<Header>().AddConstructor<
          PositionHeader>();
      return tid;
    }

    TypeId PositionHeader::GetInstanceTypeId() const
    {
      return GetTypeId();
    }

    uint32_t PositionHeader::GetSerializedSize() const
    {
      return 61;
    }

    void PositionHeader::Serialize(Buffer::Iterator i) const
    {
      i.WriteU64(m_dstPosx);
      i.WriteU64(m_dstPosy);
      i.WriteU32(m_updated);
      i.WriteU64(m_recPosx);
      i.WriteU64(m_recPosy);
      i.WriteU8(m_inRec);
      i.WriteU64(m_lastPosx);
      i.WriteU64(m_lastPosy);

      i.WriteU64(m_bufferRemaining);
    }

    uint32_t PositionHeader::Deserialize(Buffer::Iterator start)
    {
      Buffer::Iterator i = start;
      m_dstPosx = i.ReadU64();
      m_dstPosy = i.ReadU64();
      m_updated = i.ReadU32();
      m_recPosx = i.ReadU64();
      m_recPosy = i.ReadU64();
      m_inRec = i.ReadU8();
      m_lastPosx = i.ReadU64();
      m_lastPosy = i.ReadU64();
      m_bufferRemaining = i.ReadU64();

      uint32_t dist = i.GetDistanceFrom(start);
      // printf("pos deserialise\n");
      NS_ASSERT(dist == GetSerializedSize());
      return dist;
    }

    void PositionHeader::Print(std::ostream &os) const
    {
      os << " PositionX: " << m_dstPosx << " PositionY: " << m_dstPosy
         << " Updated: " << m_updated << " RecPositionX: " << m_recPosx
         << " RecPositionY: " << m_recPosy << " inRec: " << (int) m_inRec
         << " LastPositionX: " << m_lastPosx << " LastPositionY: "
         << m_lastPosy;
    }

    std::ostream &
    operator<<(std::ostream & os, PositionHeader const & h)
    {
      h.Print(os);
      return os;
    }

    bool PositionHeader::operator==(PositionHeader const & o) const
    {
      return (m_dstPosx == o.m_dstPosx && m_dstPosy == m_dstPosy
              && m_updated == o.m_updated && m_recPosx == o.m_recPosx
              && m_recPosy == o.m_recPosy && m_inRec == o.m_inRec
              && m_lastPosx == o.m_lastPosx && m_lastPosy == o.m_lastPosy);
    }

    //-----------------------------------------------------------------------------
    // SREQ
    //-----------------------------------------------------------------------------
    NS_OBJECT_ENSURE_REGISTERED (SREQHeader);

    TypeId SREQHeader::GetTypeId()
    {
      static TypeId tid =
        TypeId("ns3::gpsr::SREQHeader").SetParent<Header>().AddConstructor<
          SREQHeader>();
      return tid;
    }

    TypeId SREQHeader::GetInstanceTypeId() const
    {
      return GetTypeId();
    }

    uint32_t SREQHeader::GetSerializedSize() const
    {
      return 36;
    }

    void SREQHeader::Serialize(Buffer::Iterator i) const
    {
      i.WriteU64(m_zoneCentrex);
      i.WriteU64(m_zoneCentrey);
      i.WriteU64(m_zoneRadius);
      i.WriteU64(m_zoneExpiry);
      i.WriteU32(m_requester);
    }

    uint32_t SREQHeader::Deserialize(Buffer::Iterator start)
    {
      Buffer::Iterator i = start;

      m_zoneCentrex = i.ReadU64();
      m_zoneCentrey = i.ReadU64();
      m_zoneRadius = i.ReadU64();
      m_zoneExpiry = i.ReadU64();
      m_requester = i.ReadU32 ();

      uint32_t dist = i.GetDistanceFrom(start);
      // printf("cong deserialise\n");
      NS_ASSERT(dist == GetSerializedSize());
      return dist;
    }

    void SREQHeader::Print(std::ostream &os) const {}

    std::ostream &
    operator<<(std::ostream & os, SREQHeader const & h)
    {
      return os;
    }

    bool SREQHeader::operator==(SREQHeader const & o) const
    {
      return (true);
    }

    //-----------------------------------------------------------------------------
    // CONG
    //-----------------------------------------------------------------------------
    NS_OBJECT_ENSURE_REGISTERED (CONGHeader);

    TypeId CONGHeader::GetTypeId()
    {
      static TypeId tid =
        TypeId("ns3::gpsr::CONGHeader").SetParent<Header>().AddConstructor<
          CONGHeader>();
      return tid;
    }

    TypeId CONGHeader::GetInstanceTypeId() const
    {
      return GetTypeId();
    }

    uint32_t CONGHeader::GetSerializedSize() const
    {
      return 0;
    }

    void CONGHeader::Serialize(Buffer::Iterator i) const {}

    uint32_t CONGHeader::Deserialize(Buffer::Iterator start)
    {
      Buffer::Iterator i = start;
      uint32_t dist = i.GetDistanceFrom(start);
      // printf("cong deserialise\n");
      NS_ASSERT(dist == GetSerializedSize());
      return dist;
    }

    void CONGHeader::Print(std::ostream &os) const {}

    std::ostream &
    operator<<(std::ostream & os, CONGHeader const & h)
    {
      return os;
    }

    bool CONGHeader::operator==(CONGHeader const & o) const
    {
      return (true);
    }

  }
}
