/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#ifndef GPSRPACKET_H
#define GPSRPACKET_H

#include <iostream>
#include "ns3/header.h"
#include "ns3/enum.h"
#include "ns3/ipv4-address.h"
#include <map>
#include "ns3/nstime.h"
#include "ns3/vector.h"

namespace ns3
{
  namespace gpsr
  {

    enum MessageType
      {
        GPSRTYPE_HELLO = 1,         //!< GPSRTYPE_HELLO
        GPSRTYPE_POS = 2,            //!< GPSRTYPE_POS
        GPSRTYPE_SREQ = 3,
        GPSRTYPE_CONG = 4,
      };

    // for noting the info of cluster members
    struct MemberData
    {
      int64_t x;
      int64_t y;
      int64_t lastSeen;
      uint8_t remainingBufferPercentage;
    };

    /**
     * \ingroup gpsr
     * \brief GPSR types
     */
    class TypeHeader: public Header
    {
    public:
      /// c-tor
      TypeHeader(MessageType t);

      ///\name Header serialization/deserialization
      //\{
      static TypeId GetTypeId();
      TypeId GetInstanceTypeId() const;
      uint32_t GetSerializedSize() const;
      void Serialize(Buffer::Iterator start) const;
      uint32_t Deserialize(Buffer::Iterator start);
      void Print(std::ostream &os) const;
      //\}

      /// Return type
      MessageType Get() const
      {
        return m_type;
      }
      /// Check that type if valid
      bool IsValid() const
      {
        return m_valid; //FIXME that way it wont work
      }
      bool operator==(TypeHeader const & o) const;
    private:
      MessageType m_type;
      bool m_valid;
    };

    std::ostream & operator<<(std::ostream & os, TypeHeader const & h);

    class HelloHeader: public Header
    {
    public:
      /// c-tor
      HelloHeader(uint64_t originPosx = 0, uint64_t originPosy = 0,
                  int64_t originVelx = 0, int64_t originVely = 0);

      ///\name Header serialization/deserialization
      //\{
      static TypeId GetTypeId();
      TypeId GetInstanceTypeId() const;
      uint32_t GetSerializedSize() const;
      void Serialize(Buffer::Iterator start) const;
      uint32_t Deserialize(Buffer::Iterator start);
      void Print(std::ostream &os) const;
      //\}

      ///\name Fields
      //\{
      void SetOriginPosx(uint64_t posx)
      {
        m_originPosx = posx;
      }
      uint64_t GetOriginPosx() const
      {
        return m_originPosx;
      }
      void SetOriginPosy(uint64_t posy)
      {
        m_originPosy = posy;
      }
      uint64_t GetOriginPosy() const
      {
        return m_originPosy;
      }
      void SetOriginVelx(int64_t velx)
      {
        m_originVelx = velx;
      }
      int64_t GetOriginVelx() const
      {
        return m_originVelx;
      }
      void SetOriginVely(int64_t vely)
      {
        m_originVely = vely;
      }
      int64_t GetOriginVely() const
      {
        return m_originVely;
      }
      // new methods for augmented hello
      // centre of surveillance zone
      void SetZoneCentrex(int64_t x) { m_zoneCentrex = x; }
      int64_t GetZoneCentrex() const { return m_zoneCentrex; }
      void SetZoneCentrey(int64_t y) { m_zoneCentrey = y; }
      int64_t GetZoneCentrey() const { return m_zoneCentrey; }
      // radius
      void SetZoneRadius(int64_t r)  { m_zoneRadius = r; }
      int64_t GetZoneRadius() const  { return m_zoneRadius; }
      // expiry
      void SetZoneExpiry(int64_t e)  { m_zoneExpiry = e; }
      int64_t GetZoneExpiry() const  { return m_zoneExpiry; }
      // tracking of ids of nodes thought to be in cluster
      // only nodes seen in previous n minutes are counted as 'in'
      void AddClusterMember(uint32_t peer, MemberData md);
      std::map<uint32_t, MemberData> GetClusterMembers ()
      {
        return m_clusterMembers;
      }
      // number of stored cluster members, including self
      uint32_t GetClusterSize () const
      {
        return (uint32_t) m_clusterMembers.size ();
      }
      //\}

      bool operator==(HelloHeader const & o) const;
    private:
      uint64_t m_originPosx;          ///< Originator Position x
      uint64_t m_originPosy;          ///< Originator Position x
      int64_t m_originVelx;          ///< Originator Position x
      int64_t m_originVely;          ///< Originator Position x

      int64_t m_zoneCentrex;
      int64_t m_zoneCentrey;
      int64_t m_zoneRadius;
      int64_t m_zoneExpiry;
      std::vector<uint32_t> m_dests;
      std::map<uint32_t, MemberData> m_clusterMembers;
    };

    std::ostream & operator<<(std::ostream & os, HelloHeader const &);

    class PositionHeader: public Header
    {
    public:
      /// c-tor
      PositionHeader(uint64_t dstPosx = 0, uint64_t dstPosy = 0,
                     uint32_t updated = 0, uint64_t recPosx = 0, uint64_t recPosy = 0,
                     uint8_t inRec = 0, uint64_t lastPosx = 0, uint64_t lastPosy = 0);

      ///\name Header serialization/deserialization
      //\{
      static TypeId GetTypeId();
      TypeId GetInstanceTypeId() const;
      uint32_t GetSerializedSize() const;
      void Serialize(Buffer::Iterator start) const;
      uint32_t Deserialize(Buffer::Iterator start);
      void Print(std::ostream &os) const;
      //\}

      ///\name Fields
      //\{
      void SetDstPosx(uint64_t posx)
      {
        m_dstPosx = posx;
      }
      uint64_t GetDstPosx() const
      {
        return m_dstPosx;
      }
      void SetDstPosy(uint64_t posy)
      {
        m_dstPosy = posy;
      }
      uint64_t GetDstPosy() const
      {
        return m_dstPosy;
      }
      void SetUpdated(uint32_t updated)
      {
        m_updated = updated;
      }
      uint32_t GetUpdated() const
      {
        return m_updated;
      }
      void SetRecPosx(uint64_t posx)
      {
        m_recPosx = posx;
      }
      uint64_t GetRecPosx() const
      {
        return m_recPosx;
      }
      void SetRecPosy(uint64_t posy)
      {
        m_recPosy = posy;
      }
      uint64_t GetRecPosy() const
      {
        return m_recPosy;
      }
      void SetInRec(uint8_t rec)
      {
        m_inRec = rec;
      }
      uint8_t GetInRec() const
      {
        return m_inRec;
      }
      void SetLastPosx(uint64_t posx)
      {
        m_lastPosx = posx;
      }
      uint64_t GetLastPosx() const
      {
        return m_lastPosx;
      }
      void SetLastPosy(uint64_t posy)
      {
        m_lastPosy = posy;
      }
      uint64_t GetLastPosy() const
      {
        return m_lastPosy;
      }
      void SetRemainingBuffer(uint8_t b)
      {
        m_bufferRemaining = b;
      }
      uint64_t GetRemainingBuffer() const
      {
        return m_bufferRemaining;
      }

      bool operator==(PositionHeader const & o) const;
    protected:
      uint64_t m_dstPosx;          ///< Destination Position x
      uint64_t m_dstPosy;          ///< Destination Position x
      uint32_t m_updated;          ///< Time of last update
      uint64_t m_recPosx;          ///< x of position that entered Recovery-mode
      uint64_t m_recPosy;          ///< y of position that entered Recovery-mode
      uint8_t m_inRec;          ///< 1 if in Recovery-mode, 0 otherwise
      uint64_t m_lastPosx;          ///< x of position of previous hop
      uint64_t m_lastPosy;          ///< y of position of previous hop

      uint8_t m_bufferRemaining;

    };

    std::ostream & operator<<(std::ostream & os, PositionHeader const &);

    class SREQHeader : public Header
    {
    public:
      SREQHeader() {}
      SREQHeader(uint64_t x, uint64_t y, uint64_t r,
                 uint64_t e, uint32_t requester) :
        m_zoneCentrex (x),
        m_zoneCentrey (y),
        m_zoneRadius (r),
        m_zoneExpiry (e),
        m_requester (requester)
      {
      }
      ///\name Header serialization/deserialization
      //\{
      static TypeId GetTypeId();
      TypeId GetInstanceTypeId() const;
      uint32_t GetSerializedSize() const;
      void Serialize(Buffer::Iterator start) const;
      uint32_t Deserialize(Buffer::Iterator start);
      void Print(std::ostream &os) const;
      //\}
      bool operator== (SREQHeader const & o) const;

      void SetZoneCentrex(int64_t x) { m_zoneCentrex = x; }
      int64_t GetZoneCentrex() const { return m_zoneCentrex; }
      void SetZoneCentrey(int64_t y) { m_zoneCentrey = y; }
      int64_t GetZoneCentrey() const { return m_zoneCentrey; }
      // radius
      void SetZoneRadius(int64_t r)  { m_zoneRadius = r; }
      int64_t GetZoneRadius() const  { return m_zoneRadius; }
      // expiry
      void SetZoneExpiry(int64_t e)  { m_zoneExpiry = e; }
      int64_t GetZoneExpiry() const  { return m_zoneExpiry; }
      // last update time of radius (from scaling)
      void SetZoneUpdate(int64_t u)  { m_zoneUpdate = u; }
      int64_t GetZoneUpdate() const  { return m_zoneUpdate; }
      // requester
      void SetRequester(uint32_t a)  { m_requester = a; }
      uint32_t GetRequester() const  { return m_requester; }

    private:
      uint64_t m_zoneCentrex;
      uint64_t m_zoneCentrey;
      uint64_t m_zoneRadius;
      uint64_t m_zoneExpiry;
      uint64_t m_zoneUpdate;
      uint32_t m_requester;
    };

    std::ostream & operator<<(std::ostream & os, PositionHeader const &);

    class CONGHeader : public Header
    {
    public:
      CONGHeader() {}
      ///\name Header serialization/deserialization
      //\{
      static TypeId GetTypeId();
      TypeId GetInstanceTypeId() const;
      uint32_t GetSerializedSize() const;
      void Serialize(Buffer::Iterator start) const;
      uint32_t Deserialize(Buffer::Iterator start);
      void Print(std::ostream &os) const;
      //\}
      bool operator== (CONGHeader const & o) const;
    };

  }
}
#endif /* GPSRPACKET_H */
